import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Ellipse


def plot_orbit(planet_list: list, directory: str, name: str, num: int, figsize: tuple, title = False, dpi = 200,star_color='orange'):
    """Plots planetary system

    Plots the plantary system orbits given a list of planets to be saved as a GIF.

    Args:
        planetary_list (list): List of Planet objects representing the planets in the system.
        directory (str): Path to the directory in which to save plots.
        name (str): Name of plot files.
        num (int): Number representing order of plot in .gif animation frames. 
            Follows <name>_<num>.jpg convention.
        figsize (tuple, optional): Plot figure size in units of inches formatted as (width, height). 
            Default is (8, 8)
        title (bool, optional): Binary determiner for plot title to be included in each plot. 
            Default is False.
        dpi (int, optional): Dots per inch to use when saving the plot. 
            If kernel is crashing, try reducing the DPI.
            Default is 200
        star_color (str, optional): matplotlib color to use for the star. 
            Default is 'orange'.
    """
    fig, ax = plt.subplots(figsize = figsize, layout = 'constrained')

    star = plt.Circle((0,0), 1, color = star_color)
    ax.add_patch(star)
    
    aplusrlist = []

    for planet in planet_list:

        aplusrlist.append(planet.a*(1+planet.e)+planet.r)

        if planet.e == 0:
            orb = plt.Circle((0,0), planet.a, edgecolor = 'black', facecolor = 'None')

        else:
            orb = Ellipse((planet.a*planet.e*np.cos(planet.w), planet.a*planet.e*np.sin(planet.w)), 2*planet.a, 2*planet.a*np.sqrt(1-planet.e**2), angle = planet.w*180/np.pi, edgecolor = 'black', facecolor = 'None')

        ax.add_patch(orb)

        planetcircle = plt.Circle((planet.x, planet.y), planet.r, color = planet.color)
        ax.add_patch(planetcircle)

    axislim = max(aplusrlist)*1.1

    ax.set_xlim(-axislim, axislim)
    ax.set_ylim(-axislim, axislim)
    ax.set_aspect('equal')
    ax.axis('off')
    if title:
        ax.set_title(name, fontsize = 20)

    fig.savefig(directory+'/'+name+'_'+str(num)+'.jpg', dpi = dpi)

    plt.close()

def get_star_color(bp_rp):
    """Sets color of star for plots

    Defines the color of the star when passed data queried from Exoplanet Archive given BP/RP from Gaia DR2.

    Args:
        bp_rp (float): Color value pulled from Gaia DR2 with listed Exoplanet Archive Gaia_id using BP/RP.

    Returns:
        str: hexcode color string
    """
    colors = [
        {"min": -3, "max": -0.26,  "color": '#0000ff'},
        {"min": -0.25, "max": -0.038, "color": '#699bff'},
        {"min": -0.037, "max": 0.327, "color": '#00eded'},
        {"min": 0.326, "max": 0.767,  "color": "#F9E400"},
        {"min": 0.768, "max": 0.984, "color": '#ffbe0d'},
        {"min": 0.983, "max": 1.85, "color": '#ff801f'},
        {"min": 1.85, "max": 7.0, "color": '#d40b0b'}
    ]
    for i in range(len(colors)):
        if bp_rp>colors[i]['min'] and bp_rp<colors[i]['max']:
            return colors[i]['color']
    if bp_rp<-3: 
        return '#5b7cff'
    if bp_rp>7.0:
        return '#ffa448'


import os
from Orbeez import orbeez

def test_make_orbit_gif(directory: str='tests/end-to-end-tests/'):
    """
    End to end test for making .gif animation from user generated entry

    Args:
        directory (str, optional): Path to the directory in which to save the resulting .gif animation.
            Default is end-to-end-tests folder.
    """

    a_list = [1, 2, 3]
    p_list = [1, 2, 3]
    r_list = [1, 1, 1]

    orbeez.make_orbit_gif(a_list, p_list, r_list, directory, 'planet')
    assert os.path.isfile(directory+'planet.gif')


test_make_orbit_gif()


import numpy as np
from scipy.optimize import fsolve

class Planet:
    """A class that stores the information for a planet.

    This class keeps track of the orbital distances, relative radii, and the color of each planet that should be plotted in a given system.

    Attributes:
        a (float): semi-major axis for planet in units of stellar radii
        p (float): period for planet in any unit consistent with other planetary orbits
        r (float): radius for planet in units of stellar radii
        e (float): eccentricity for planet
        w (float): argument of periastron for planet in units of radians
        color (str): color which the planet itself (not the orbital circle) will be plotted with
        x (float): x-coordinate defining position of planet along orbit 
        y (float): y-coordinate defining position of planet along orbit
    """
    def __init__(self, a, p, r, e, w, color, planet_r_scale = 25):
        """
        Args:
            a (float): semi-major axis for planet in units of stellar radii
            p (float): period for planet in any unit consistent with other planetary orbits
            r (float): radius for planet in units of stellar radii
            e (float): eccentricity for planet
            w (float): argument of periastron for planet in units of radians
            color (str): color which the planet itself (not the orbital circle) will be plotted with
            planet_r_scale (float, optional): Factor by which to scale up the planet radii. Default is 25.
        """

        self.a = a
        self.p = p
        self.r = r*planet_r_scale
        self.e = e
        self.w = w
        if color is None:
            self.color = 'black'
        else:
            self.color = color
        self.x  = 0
        self.y = self.a

        if self.e > 0:

            focc = -np.pi/2 - self.w
            Eocc = np.arctan2(np.sqrt(1-self.e**2)*np.sin(focc), self.e+np.cos(focc))
            self.Tp = -self.p/(2*np.pi) * (Eocc - self.e*np.sin(Eocc))
        
    def update_pos(self, t):
        """
        Update the position of the planet along its orbit given a timestep.

        Args:
            t (float): timestep defining how far to move planet along orbit
        """
      
        if self.e == 0:

            theta = t/self.p*2*np.pi + np.pi/2
            self.x = self.a*np.cos(theta)
            self.y = self.a*np.sin(theta)

        else:

            E = self.solve_for_E(t)
            f = np.arctan2(np.sqrt(1-self.e**2)*np.sin(E), np.cos(E)-self.e)
            d = self.a*(1-self.e**2)/(1+self.e*np.cos(f))
            self.x = -d*np.cos(self.w+f)
            self.y = -d*np.sin(self.w+f)

    def solve_for_E(self, t):
        """
        Solve or the eccentric anomaly of the planet at a given time t.

        Args:
            t (float): time at which to calculate the eccentric anomaly
        """

        return fsolve(lambda x: x - self.e * np.sin(x) - 2*np.pi / self.p * (t-self.Tp), 2*np.pi / self.p * (t-self.Tp))[0]

        



import os
from Orbeez import orbeez

def test_gif_from_archive(directory: str='tests/end-to-end-tests/'):
    """
    End to end test for making .gif animation from Exoplanet Archive entry

    Args:
        directory (str, optional): Path to the directory in which to save the resulting .gif animation.
            Default is end-to-end-tests folder.
    """

    orbeez.gif_from_archive('TRAPPIST-1', directory)
    assert os.path.isfile(directory+'TRAPPIST-1.gif')



test_gif_from_archive()




import os
import numpy as np
from Orbeez.planet import Planet
from PIL import Image
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive
from astropy import units as u
from Orbeez.orbitplot import plot_orbit, get_star_color
import tqdm
from astroquery.gaia import Gaia



def make_orbit_gif(a_list, p_list, r_list, directory, name, e_list = None, w_list = None, figsize=(8,8), num_periods = 1, gif_duration = 10.0, color_list=None, star_color='orange', num_frames=100, title = False, dpi = 200, planet_r_scale = 25):
    """Makes a .gif animation of the orbits of the input planetary system.

    Generates a .gif animation of the orbits of planets in a system architecture defined by the user.

    Args:
        a_list (array_like): List of semimajor axis values for the planets in the system, in units of stellar radii.
        p_list (array_like): List of orbital period values for the planets in the system, in any consistent units.
        r_list (array_like): List of planetary radii values for the planets in the system, in units of stellar radii.
        directory (str): Path to the directory in which to save the resulting .gif animation.
        name (str): Name of the resulting .gif animation.
        e_list (array_like, optional): List of eccentricity values for the planets in the system. If None, default
            initializes the eccentricities of all planets to 0. Default is None.
        w_list (array_like, optional): List of arguments of periastron for the planets in the system, in units of
            radians. If None, default initializes the arguments of periastron of all planets to pi/2. Default is None.
        figsize (tuple, optional): Size of the .gif animation in units of inches. Formatted as (width, height).
            Default is (8,8).
        num_periods (int, optional): Number of periods of the outermost planet to animate. Default is 1.
        gif_duration (float, optional): Duration of the whole .gif animation in seconds. Default is 10 seconds.
        color_list (array_like, optional): List of matplotlib colors to loop through when plotting the planets.
            Default is None, which sets the planets to be black.
        star_color (str, optional): matplotlib color to use for the star. Default is orange.
        num_frames (int, optional): Number of frames to use in the .gif animation. More frames will make the
            animation more smooth, but will slow down the creation process. Too many frames may cause the kernel
            to crash when making the .gif. Default is 100.
        title (bool, optional): Whether or not to include the name as a title above the animation. Default is False.
        dpi (int, optional): Dots per inch to use when saving the frames. If the kernel is crashing, try reducing the
            dpi. Default is 200.
        planet_r_scale (float, optional): Factor by which to scale up the planet radii. Default is 25.
    """

    if e_list is None or w_list is None:

        e_list = [0]*len(a_list)
        w_list = [np.pi/2]*len(a_list)
        

    if not len(a_list) == len(p_list) == len(r_list) == len(e_list) == len(w_list):
        print('Planet arrays not same length')
        return 
    
    if np.any(np.isnan(r_list)):
        print('Query returned some nan planetary radii, setting radii to default value of 0.01')
        indices = np.where(np.isnan(r_list))[0]
        for i in indices:
            r_list[i] = 0.01

    if color_list is None:
        color_list = [None] * len(a_list)
    elif len(color_list) < len(a_list):
        color_list *= int(np.ceil(len(a_list)/len(color_list)))
        

    p_list = np.array(p_list)/max(p_list)

    if np.min(p_list)*num_frames < 16:
        num_frames = int(np.ceil(16/np.min(p_list)))

    planet_list = []

    for i in range(len(a_list)):
        entry = Planet(a_list[i], p_list[i], r_list[i], e_list[i], w_list[i], color_list[i], planet_r_scale = planet_r_scale)
        planet_list.append(entry)

    print('Generating images...')
    for j in tqdm.tqdm(range(num_frames)):
        for planet in planet_list:
            planet.update_pos(j/num_frames*num_periods)
        plot_orbit(planet_list, directory, name, j, figsize, title = title, dpi = dpi, star_color=star_color)

    print('Stitching frames...')
    frames = [Image.open(directory+'/'+name+'_'+str(i)+'.jpg') for i in tqdm.tqdm(range(num_frames))]

    frame_1 = frames[0]
    frame_1.save(directory+'/'+name+'.gif', format='GIF', append_images=frames, save_all=True, duration=gif_duration/num_frames*1000, loop=0)

    print('Deleting images...')
    for j in tqdm.tqdm(range(num_frames)):
        os.remove(directory+'/'+name+'_'+str(j)+'.jpg')


def gif_from_archive(system_name, directory, figsize=(8,8), num_periods = 1, gif_duration = 10.0, color_list=None, num_frames=100, title = False, dpi = 200, planet_r_scale = 25):
     
    """Generates gif of exoplanet system with user entered name from NASA Exoplanet Archive.

    Generates a .gif animation of the orbits of planets in a system architecture defined by data pulled from the Exoplanet Archive chosen by the user by passing the name of a host star.

    Args:
        system_name (str): Name of the exoplanet system as found in NASA exoplanet Archive.
        directory (str): Directory where you would like the gif to be saved.
        figsize (tuple, optional): Size of the .gif animation in units of inches. Formatted as (width, height).
            Default is (8,8).
        num_periods (int, optional): Number of periods of the outermost planet to animate. Default is 1.
        gif_duration (float, optional): Duration of the whole .gif animation in seconds. Default is 10 seconds.
        color_list (list, optional): List of matplotlib colors to loop through when plotting the planets.
            Default is None, which sets the planets to be black.
        num_frames (int, optional): Number of frames to use in the .gif animation. More frames will make the
            animation more smooth, but will slow down the creation process. Too many frames may cause the kernel
            to crash when making the .gif. Default is 100.
        title (bool, optional): Whether or not to include the name as a title above the animation. Default is False.
        dpi (int, optional): Dots per inch to use when saving the frames. If the kernel is crashing, try reducing the
            dpi. Default is 200.
        planet_r_scale (float, optional): Factor by which to scale up the planet radii. Default is 25.
    """
    
    data = NasaExoplanetArchive.query_criteria(
        table="ps", 
        select="pl_name, pl_orbsmax, pl_orbper, pl_radj, pl_orbeccen, pl_orblper, st_rad, gaia_id",
        where="hostname='{}' AND default_flag=1".format(system_name),
    )

    data.sort('pl_orbper')

    a_list = (data['pl_orbsmax'].to(u.Rsun)/data['st_rad']).value
    p_list = data['pl_orbper'].value
    r_list = (data['pl_radj'].to(u.Rsun)/data['st_rad']).value
    e_list = (data['pl_orbeccen']).value
    w_list = (data['pl_orblper']).value * np.pi/180

    i = np.where(np.isnan(e_list))[0]
    e_list[i] = 0
    w_list[i] = np.pi/2

    if np.any(np.isnan(w_list)):
        print('Some nan arguments of periastron, setting to pi/2.')
        i = np.where(np.isnan(w_list))[0]
        w_list[i] = np.pi/2

    if np.all(np.isnan(a_list)):
        print('No semi-major axes available.')
        return
    
    if np.all(np.isnan(p_list)):
        print('No periods available.')
        return
    
    if np.any(np.isnan(a_list) & np.isnan(p_list)):
        print('One planet has no semi-major axis and no period.')
        return

    if np.any(np.isnan(a_list)):
        i = np.where(~np.isnan(a_list))[0][0]
        p0 = p_list[i]
        a0 = a_list[i]
        j = np.where(np.isnan(a_list))[0]
        a_list[j] = (p_list[j]/p0)**(2/3) * a0

    if np.any(np.isnan(p_list)):
        i = np.where(~np.isnan(p_list))[0][0]
        p0 = p_list[i]
        a0 = a_list[i]
        j = np.where(np.isnan(p_list))[0]
        p_list[j] = (a_list[j]/a0)**(3/2) * p0

    gaiaid=data['gaia_id'][0].split()[2]
    query = f"SELECT bp_rp FROM gaiadr2.gaia_source WHERE source_id = {gaiaid}"
    job = Gaia.launch_job(query)
    bp_rp= job.get_data()['bp_rp'][0]

    star_color=get_star_color(bp_rp)
    

    make_orbit_gif(a_list, p_list, r_list, directory=directory, name=system_name, e_list = e_list, w_list = w_list, figsize=figsize, num_periods=num_periods, gif_duration=gif_duration, color_list=color_list, star_color=star_color, num_frames=num_frames, title=title, dpi=dpi, planet_r_scale = planet_r_scale)




import os
from Orbeez import orbitplot
from Orbeez.planet import Planet
import numpy as np


def test_plot_orbit():
    """
    Tests whether plot_orbit generates a .png file of a single GIF frame
    """
    p = Planet(1, 1, 1, 0, np.pi/2, 'blue')

    orbitplot.plot_orbit([p], directory='tests/', name='test_plot', num=1, figsize=(8, 8))

    assert os.path.isfile('tests/test_plot_1.jpg')



from Orbeez.planet import Planet
import numpy as np
import pytest


def test_planet():
    """
    Tests the proper instantiation of a Planet object and that it integrates correctly
    """

    p = Planet(1, 1, 1, 0, np.pi/2, 'blue')
    p.update_pos(t=1)

    assert p.x == pytest.approx(1*np.cos(1/1*2*np.pi + np.pi/2), abs=1e-3)
    assert p.y == pytest.approx(1*np.sin(1/1*2*np.pi + np.pi/2), abs=1e-3)
